/*
 *	File:		PPPPPPPP_CocoaView.m
 *
 *	Version:	1.0
 *
 *	Created:	2/22/14
 *
 *	Template by Alex Kenis 2014
 */

#import "PPPPPPPP_CocoaView.h"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Re-list the parameters that we set up in our AUSource header file
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
enum
{
	kParam_One =0,
    kParam_Two =1,
    kParam_Three =2,
	kNumberOfParameters=3
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Setup callback listeners for mouse actions
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark - Listener Callback Dispatcher

// This listener responds to parameter changes, gestures, and property notifications
void ParameterListenerDispatcher (void *inRefCon, void *inObject, const AudioUnitParameter *inParameter, Float32 inValue)
{
	PPPPPPPP_CocoaView *SELF = (__bridge PPPPPPPP_CocoaView *)inRefCon;
    
    [SELF _parameterListener:inObject parameter:inParameter value:inValue];
}

NSString *PPPPPPPP_GestureSliderMouseDownNotification = @"CAGestureSliderMouseDownNotification"; // notification broadcast by the view when the user has started a gesture
NSString *PPPPPPPP_GestureSliderMouseUpNotification = @"CAGestureSliderMouseUpNotification"; // notification broadcast by the view when the user has finished a gesture

@implementation PPPPPPPP_GestureSlider

/*	We create our own custom subclass of NSSlider so we can do begin/end gesture notification
 We cannot override mouseUp: because it will never be called. Instead we do a clever trick in mouseDown to send mouseUp notifications */
- (void)mouseDown:(NSEvent *)inEvent
{
	[[NSNotificationCenter defaultCenter] postNotificationName: PPPPPPPP_GestureSliderMouseDownNotification object: self];
	
	[super mouseDown: inEvent];	// this call does not return until mouse tracking is complete
    // once tracking is done, we know the mouse has been released, so we can send our mouseup notification
    
	[[NSNotificationCenter defaultCenter] postNotificationName: PPPPPPPP_GestureSliderMouseUpNotification object: self];
}

@end

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Release listeners
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@implementation PPPPPPPP_CocoaView

- (void)dealloc
{
    [self _removeListeners];
    [super dealloc]; // you can get id of this is you are using ARC
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Set an array of our controls
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark - Public Functions

- (void)setAU:(AudioUnit)inAU
{
	// remove previous listeners
	if (mAU) [self _removeListeners];
	mAU = inAU;
    
    // 1 Gain
   	mParameter[0].mAudioUnit = inAU; // the audiounit
	mParameter[0].mParameterID = kParam_One; // gain parameter
	mParameter[0].mScope = kAudioUnitScope_Global;
	mParameter[0].mElement = 0; // we have only one element
    
    // 2 Tone
   	mParameter[1].mAudioUnit = inAU;
	mParameter[1].mParameterID = kParam_Two;
	mParameter[1].mScope = kAudioUnitScope_Global;
	mParameter[1].mElement = 0;
    
    // 3 Vol
   	mParameter[2].mAudioUnit = inAU;
	mParameter[2].mParameterID = kParam_Three;
	mParameter[2].mScope = kAudioUnitScope_Global;
	mParameter[2].mElement = 0;
    
	// add new listeners
	[self _addListeners];
	
	// initial setup
	[self _synchronizeUIWithParameterValues];
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Set up our interface actions and tie them to the listeners and parameter array that
//      we setup in the above sections/  Both the slider and the value readout text
//      field send values, so we set them both to the same IBAction
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark - Interface Actions

// Gain
- (IBAction)iaParam1Changed:(id)sender
{
    float floatValue = [sender floatValue];
	NSAssert(	AUParameterSet(mParameterListener, sender, &mParameter[0], (Float32)floatValue, 0) == noErr,
             @"[PPPPPPPP_CocoaView iaParam1Changed:] AUParameterSet()");
    if (sender == uiParam1Slider) {
        [uiParam1TextField setFloatValue:floatValue];  // text field
    } else {
        [uiParam1Slider setFloatValue:floatValue];      // slider control
    }
}

// Tone
- (IBAction)iaParam2Changed:(id)sender
{
    float floatValue = [sender floatValue];
	NSAssert(	AUParameterSet(mParameterListener, sender, &mParameter[1], (Float32)floatValue, 0) == noErr,
             @"[PPPPPPPP_CocoaView iaParam2Changed:] AUParameterSet()");
    if (sender == uiParam2Slider) {
        [uiParam2TextField setFloatValue:floatValue];
    } else {
        [uiParam2Slider setFloatValue:floatValue];
    }
}

//Vol
- (IBAction)iaParam3Changed:(id)sender
{
    float floatValue = [sender floatValue];
	NSAssert(	AUParameterSet(mParameterListener, sender, &mParameter[2], (Float32)floatValue, 0) == noErr,
             @"[PPPPPPPP_CocoaView iaParam3Changed:] AUParameterSet()");
    if (sender == uiParam3Slider) {
        [uiParam3TextField setFloatValue:floatValue];
    } else {
        [uiParam3Slider setFloatValue:floatValue];
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Handle mouse actions and tie the notifications to each slider
// This routine is called when the user has clicked on the slider.
// We need to send a begin parameter change gesture to alert hosts
//  that the parameter may be changing value
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark - Notifications

// This routine is called when the user has clicked on the slider. We need to send a begin parameter change gesture to alert hosts that the parameter may be changing value

// ~~~~ MOUSE DOWN ~~~~ //
- (void)handleMouseDown: (NSNotification *) aNotification
{
    // Gain
	if ([aNotification object] == uiParam1Slider) {
		AudioUnitEvent event;
		event.mArgument.mParameter = mParameter[0]; // gain array number
		event.mEventType = kAudioUnitEvent_BeginParameterChangeGesture;
		
		AUEventListenerNotify(NULL, (__bridge void *)(self), &event);
	}
    
    // Tone
    if ([aNotification object] == uiParam2Slider) {
		AudioUnitEvent event;
		event.mArgument.mParameter = mParameter[1]; // tone array number
		event.mEventType = kAudioUnitEvent_BeginParameterChangeGesture;
		
		AUEventListenerNotify(NULL, (__bridge void *)(self), &event);
	}
    
    // Vol
    if ([aNotification object] == uiParam3Slider) {
		AudioUnitEvent event;
		event.mArgument.mParameter = mParameter[2]; // vol array number
		event.mEventType = kAudioUnitEvent_BeginParameterChangeGesture;
		
		AUEventListenerNotify(NULL, (__bridge void *)(self), &event);
	}
}

// ~~~~ MOUSE UP ~~~~ //
-(void)handleMouseUp: (NSNotification *) aNotification
{
    //Gain
	if ([aNotification object] == uiParam1Slider) {
		AudioUnitEvent event;
		event.mArgument.mParameter = mParameter[0];
		event.mEventType = kAudioUnitEvent_EndParameterChangeGesture;
        
		AUEventListenerNotify(NULL, (__bridge void *)(self), &event);
	}
    
    // Tone
    if ([aNotification object] == uiParam2Slider) {
		AudioUnitEvent event;
		event.mArgument.mParameter = mParameter[1];
		event.mEventType = kAudioUnitEvent_EndParameterChangeGesture;
        
		AUEventListenerNotify(NULL, (__bridge void *)(self), &event);
	}
    
    // Vol
    if ([aNotification object] == uiParam3Slider) {
		AudioUnitEvent event;
		event.mArgument.mParameter = mParameter[2];
		event.mEventType = kAudioUnitEvent_EndParameterChangeGesture;
        
		AUEventListenerNotify(NULL, (__bridge void *)(self), &event);
	}
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Create listeners for the combines actions that we tied together above
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark - Private Functions

- (void)_addListeners
{
	NSAssert (	AUListenerCreate(	ParameterListenerDispatcher, self,
                                 CFRunLoopGetCurrent(), kCFRunLoopDefaultMode, 0.100,
                                 &mParameterListener	) == noErr,
              @"[PPPPPPPP_CocoaView _addListeners] AUListenerCreate()");
	
    int i;
    for (i = 0; i < kNumberOfParameters; ++i) {
        mParameter[i].mAudioUnit = mAU;
        NSAssert (	AUListenerAddParameter (mParameterListener, NULL, &mParameter[i]) == noErr,
                  @"[PPPPPPPP_CocoaView _addListeners] AUListenerAddParameter()");
    }
    
    // Gain
   	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleMouseDown:) name:PPPPPPPP_GestureSliderMouseDownNotification object:uiParam1Slider];    // Mouse-down gain
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleMouseUp:) name:PPPPPPPP_GestureSliderMouseUpNotification object:uiParam1Slider];    //Mouse-up gain
    
    // Tone
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleMouseDown:) name:PPPPPPPP_GestureSliderMouseDownNotification object:uiParam2Slider];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleMouseUp:) name:PPPPPPPP_GestureSliderMouseUpNotification object:uiParam2Slider];
    
    // Vol
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleMouseDown:) name:PPPPPPPP_GestureSliderMouseDownNotification object:uiParam3Slider];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleMouseUp:) name:PPPPPPPP_GestureSliderMouseUpNotification object:uiParam3Slider];
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Put away listeners when we are done with them
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)_removeListeners
{
    int i;
    for (i = 0; i < kNumberOfParameters; ++i) {
        NSAssert (	AUListenerRemoveParameter(mParameterListener, NULL, &mParameter[i]) == noErr,
                  @"[PPPPPPPP_CocoaView _removeListeners] AUListenerRemoveParameter()");
    }
    
	NSAssert (	AUListenerDispose(mParameterListener) == noErr,
              @"[PPPPPPPP_CocoaView _removeListeners] AUListenerDispose()");
    
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

- (void)_synchronizeUIWithParameterValues
{
	Float32 value;
    int i;
    
    for (i = 0; i < kNumberOfParameters; ++i) {
        NSAssert (	AudioUnitGetParameter(mAU, mParameter[i].mParameterID, kAudioUnitScope_Global, 0, &value) == noErr,
                  @"[PPPPPPPP_CocoaView synchronizeUIWithParameterValues] (x.1)");
        NSAssert (	AUParameterSet (mParameterListener, self, &mParameter[i], value, 0) == noErr,
                  @"[PPPPPPPP_CocoaView synchronizeUIWithParameterValues] (x.2)");
        NSAssert (	AUParameterListenerNotify (mParameterListener, self, &mParameter[i]) == noErr,
                  @"[PPPPPPPP_CocoaView synchronizeUIWithParameterValues] (x.3)");
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Format and send out our UI values for both the sliders and the text readout fields
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark - Listener Callback Dispatchee

- (void)_parameterListener:(void *)inObject parameter:(const AudioUnitParameter *)inParameter value:(Float32)inValue
{
    // inObject ignored in this case.
    
	switch (inParameter->mParameterID)
    {
            //Gain
        case kParam_One:
            [uiParam1Slider setFloatValue:inValue]; // gain slider parameter
            [uiParam1TextField setStringValue:[[NSNumber numberWithFloat:inValue] stringValue]]; // gain text parameter
            break;
            
            // Tone
		case kParam_Two:
            [uiParam2Slider setFloatValue:inValue];
            [uiParam2TextField setStringValue:[[NSNumber numberWithFloat:inValue] stringValue]];
            break;
            
            // Vol
        case kParam_Three:
            [uiParam3Slider setFloatValue:inValue];
            [uiParam3TextField setStringValue:[[NSNumber numberWithFloat:inValue] stringValue]];
            break;
	}
}
@end
